import { CommandHandler, ICommandHandler } from '@nestjs/cqrs';
import { CreateAdCommand } from './create-ad.command';
import { Inject } from '@nestjs/common';
import {
  AD_REPOSITORY,
  AD_CARPOOL_ROUTE_PROVIDER,
} from '@modules/ad/ad.di-tokens';
import { AdEntity } from '@modules/ad/core/domain/ad.entity';
import { AdRepositoryPort } from '../../ports/ad.repository.port';
import { AggregateID, ConflictException } from '@mobicoop/ddd-library';
import { AdAlreadyExistsException } from '@modules/ad/core/domain/ad.errors';
import { CarpoolRouteProviderPort } from '../../ports/carpool-route-provider.port';
import { Role } from '@modules/ad/core/domain/ad.types';
import { CarpoolRoute } from '../../types/carpool-route.type';

@CommandHandler(CreateAdCommand)
export class CreateAdService implements ICommandHandler {
  constructor(
    @Inject(AD_REPOSITORY)
    private readonly repository: AdRepositoryPort,
    @Inject(AD_CARPOOL_ROUTE_PROVIDER)
    private readonly carpoolRouteProvider: CarpoolRouteProviderPort,
  ) {}

  async execute(command: CreateAdCommand): Promise<AggregateID> {
    const roles: Role[] = [];
    if (command.driver) roles.push(Role.DRIVER);
    if (command.passenger) roles.push(Role.PASSENGER);
    const carpoolRoute: CarpoolRoute = await this.carpoolRouteProvider.getBasic(
      roles,
      command.waypoints,
    );
    const ad = AdEntity.create({
      id: command.id,
      driver: command.driver,
      passenger: command.passenger,
      frequency: command.frequency,
      fromDate: command.fromDate,
      toDate: command.toDate,
      schedule: command.schedule,
      seatsProposed: command.seatsProposed,
      seatsRequested: command.seatsRequested,
      strict: command.strict,
      waypoints: command.waypoints,
      points: carpoolRoute.points,
      driverDistance: carpoolRoute.driverDistance,
      driverDuration: carpoolRoute.driverDuration,
      passengerDistance: carpoolRoute.passengerDistance,
      passengerDuration: carpoolRoute.passengerDuration,
      fwdAzimuth: carpoolRoute.fwdAzimuth,
      backAzimuth: carpoolRoute.backAzimuth,
    });

    try {
      await this.repository.insertExtra(ad, 'ad');
      return ad.id;
    } catch (error: any) {
      if (error instanceof ConflictException) {
        throw new AdAlreadyExistsException(error);
      }
      throw error;
    }
  }
}
